#define _GNU_SOURCE
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>

#include "uhumenu.h"

static int path_remove;

static int parse_item (struct item *item)
{
	struct key_value **kvp;
	char *p, *c, *e, *s;
	char dummy;

	kvp = &item->keys_values;
	p = item->file_content;
	while (p < item->file_content + item->file_size) {
		c = e = NULL;
		s = p;
		while (1) {
			if (*s == '#' && c == NULL) c = s;
			if (*s == '=' && e == NULL) e = s;
			if (*s == '\n' || *s == '\0') break;
			s++;
		}
		if (e != NULL && (c == NULL || c > e)) {
			*kvp = malloc(sizeof(struct key_value));
			(*kvp)->key = strndup(p, e - p);
			(*kvp)->value = strndup(e + 1, s - e - 1);
			kvp = &((*kvp)->next);
		}
		p = s+1;
	}
	*kvp = NULL;

	s = get_value(item, "Major");
	if (s == NULL || sscanf(s, "%d%c", &item->major, &dummy) != 1) item->major = 0;
	s = get_value(item, "Minor");
	if (s == NULL || sscanf(s, "%d%c", &item->minor, &dummy) != 1) item->minor = 0;

	s = get_value(item, "Terminal");
	item->terminal = (s != NULL && (!strcmp(s, "true") || !strcmp(s, "1")));

	item->wmlist = get_value(item, "WindowManagers");

	return 0;
}

static int file_to_item (const char *file, struct item *item)
{
	struct stat st;
	int i;
	char buf[PATH_MAX];
	const char *filename;
	int fd;

	i = stat(file, &st);
	if (i < 0) {
		fprintf(stderr, "cannot stat %s: %s\n", file, strerror(errno));
		return -1;
	}
	if (S_ISREG(st.st_mode)) {
		item->issubmenu = 0;
		filename = file;
	} else if (S_ISDIR(st.st_mode)) {
		item->issubmenu = 1;
		sprintf(buf, "%s/.directory", file);
		filename = buf;
	} else {
		fprintf(stderr, "%s is not a regular file or directory\n", file);
		return -1;
	}

	fd = open(filename, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "cannot open %s: %s\n", filename, strerror(errno));
		return -1;
	}

	if (fstat(fd, &st) < 0) {
		fprintf(stderr, "fstat: %s\n", strerror(errno));
		close(fd);
		return -1;
	}

	item->file_name = strdup(file + path_remove);
	item->file_size = st.st_size;
	item->file_content = malloc(item->file_size + 64);

	if (read(fd, item->file_content, item->file_size) != item->file_size) {
		fprintf(stderr, "read failed\n");
		close(fd);
		return -1;
	}
	memset(item->file_content + item->file_size, 0, 64);

	close(fd);

	return 0;
}

static struct dirent *readdir_nobackup (DIR *d)
{
	struct dirent *de;
	while ((de = readdir(d)) != NULL) {
		if (de->d_name[0] == '.') continue;
		if (strstr(de->d_name, ".dpkg") != NULL) continue;
		if (strstr(de->d_name, ".rpm") != NULL) continue;
		if (strstr(de->d_name, ".orig") != NULL) continue;
		if (strstr(de->d_name, ".bak") != NULL) continue;
		if (strstr(de->d_name, "~") != NULL) continue;
		return de;
	}
	return NULL;
}

static struct menu *build_menu (const char *dir)
{
	DIR *d;
	struct dirent *de;
	char buf[PATH_MAX];
	struct menu *menu;
	int i;

	d = opendir(dir);
	if (d == NULL) {
		fprintf(stderr, "opendir failed: %s\n", strerror(errno));
		return NULL;
	}
	menu = malloc(sizeof(struct menu));
	menu->nr = 0;
	while ((de = readdir_nobackup(d)) != NULL) menu->nr++;
	menu->items = malloc(menu->nr * sizeof(struct item));
	rewinddir(d);
	i = 0;
	while (i < menu->nr && (de = readdir_nobackup(d)) != NULL) {
		sprintf(buf, "%s/%s", dir, de->d_name);
		if (file_to_item(buf, &menu->items[i]) < 0) continue;
		if (parse_item(&menu->items[i]) < 0) continue;
		if (menu->items[i].issubmenu) {
			if ((menu->items[i].submenu = build_menu(buf)) == NULL) {
				continue;
			}
		}
		i++;
	}
	menu->nr = i;
	closedir(d);
	if (i == 0) {
		free(menu->items);
		free(menu);
		return NULL;
	}
	return menu;
}

struct menu *read_menu (const char *path)
{
	struct menu *menu;

	path_remove = strlen(path);

	menu = malloc(sizeof(struct menu));
	menu->nr = 1;
	menu->items = malloc(sizeof(struct item));
	if (file_to_item(path, &menu->items[0]) < 0) return NULL;
	if (!menu->items[0].issubmenu) {
		fprintf(stderr, "%s is not a directory\n", path);
		return NULL;
	}
	if (parse_item(&menu->items[0]) < 0) return NULL;
	if ((menu->items[0].submenu = build_menu(path)) == NULL) menu->nr = 0;

	return menu;
}
